from markdown.postprocessors import Postprocessor
from lxml import etree
import re

class TEIPostprocesor(Postprocessor):

    def run(self, text):
        root_tag = 'root'
        indent_unit = '  '

        tree = etree.fromstring(f'<{root_tag}>{text}</{root_tag}>')

        # Replace some html:
        for el in tree.xpath('//em'):
            el.tag = 'hi'
            el.attrib['rend'] = 'italic'

        for el in tree.xpath('//strong'):
            el.tag = 'hi'
            el.attrib['rend'] = 'bold'

        for el in tree.xpath('//h1'):
            el.tag = 'head'

        etree.indent(tree, space = indent_unit)

        new_text = etree.tostring(tree, encoding = 'unicode')
        # Remove wrapped root element and remove one level of indentation
        new_text = re.sub(rf'</?{root_tag}>\s*|(?<=\n){indent_unit}', '', new_text)

        return new_text



import re
from xml.etree import ElementTree as etree

from markdown.preprocessors import Preprocessor
from markdown.blockprocessors import BlockProcessor


class LineNumsPreproc(Preprocessor):

    RE_LINE = re.compile(r'^([0-9]+[a-z]?´?(?:[-/][0-9]*[a-z]?)?)\.(-?)\s*(.*?)(?<!\s)\s*$')

    def run(self, lines):
        new_lines = []
        prev = True # Track whether previous line has a line number or not
                    # so we can separate numbered lines in blocks
        for line in lines:
            match = self.RE_LINE.search(line)
            if match:
                n, hyph, rem = match.groups()
                n = n.replace('-','–') # Use en-dash for ranges instead of hyphen
                line_break = ' break=no' if hyph else ''
                if not prev:
                    new_lines.append('')
                new_lines.append(f'{{LINE n={n}{line_break}}}{rem}')
                prev = True
            else:
                if prev:
                    new_lines.append('')
                new_lines.append(line)
                prev = False
        return new_lines


class NumberedBlocksProcessor(BlockProcessor):

    RE_LINE_DECL = re.compile(r'^\{LINE (.+?)\}')

    def test(self, parent, block):
        return all(self.RE_LINE_DECL.search(line) for line in block.split('\n'))

    def run(self, parent, blocks):
        lines = blocks.pop(0).split('\n')
        for line in lines:
            match = self.RE_LINE_DECL.search(line)
            attribs = [attr.split('=') for attr in match.group(1).split(' ')]
            content = line[match.end():]

            line_el = etree.SubElement(parent, 'l')
            for attr in attribs:
                line_el.set(*attr)
            line_el.text = content



from markdown.inlinepatterns import InlineProcessor
from xml.etree import ElementTree as etree

RE_HETA = r'(?:[_*]\u0323?)(h\u0323?)(?:[_*]\u0323?)'

class HetaProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        el = etree.Element('em')
        el.text = m.group(1)
        return el, m.start(), m.end()



from markdown.extensions import Extension

from .divisions import DivisionsPreproc, DivisionMarkProcessor, DivisionMarkTreeproc, TrivialProcessor
from .line_nums import LineNumsPreproc, NumberedBlocksProcessor
from .milestones import RE_MIL, MilestoneProcessor
from .gaps_spaces import RE_SPACE, SpaceProcessor, RE_LINE_GAP, LineGapProcessor, RE_CHARACTER_GAP, CharacterGapProcessor
from .to_xml import TEIPostprocesor
from .misc import HetaProcessor, RE_HETA


class LeidenPlus(Extension):

    def __init__(self, **kwargs):
        self.config = {
            'strict' : [False, 'Toggle strict mode (omit all Markdown specific conventions)']
        }
        super(LeidenPlus, self).__init__(**kwargs)

    def extendMarkdown(self, md):
        configs = self.getConfigs()

        md.parser.blockprocessors.deregister('paragraph')
        md.parser.blockprocessors.deregister('olist')
        if configs['strict']:
            md.inlinePatterns.deregister('em_strong')
            md.inlinePatterns.deregister('em_strong2')

        md.preprocessors.register(DivisionsPreproc(md), 'divison_preproc', 120)
        md.preprocessors.register(LineNumsPreproc(md), 'linenums_preproc', 119)
        md.parser.blockprocessors.register(NumberedBlocksProcessor(md.parser), 'lineblocks', 120)
        md.parser.blockprocessors.register(DivisionMarkProcessor(md.parser), 'divison_marks', 119)
        md.parser.blockprocessors.register(TrivialProcessor(md.parser), 'fallback', 0)
        md.inlinePatterns.register(MilestoneProcessor(RE_MIL, md), 'milestones', 120)
        md.inlinePatterns.register(SpaceProcessor(RE_SPACE, md), 'spaces', 119)
        md.inlinePatterns.register(LineGapProcessor(RE_LINE_GAP, md), 'line_gaps', 118)
        md.inlinePatterns.register(CharacterGapProcessor(RE_CHARACTER_GAP, md), 'character_gaps', 117)
        md.inlinePatterns.register(HetaProcessor(RE_HETA, md), 'heta', 49) # After UnderscoreProcessor
        md.treeprocessors.register(DivisionMarkTreeproc(md), 'divison_treeproc', 120)
        md.postprocessors.register(TEIPostprocesor(md), 'to_xml', 0)


class LeidenEscape(Extension):

    def extendMarkdown(self, md):
        md.postprocessors.register(TEIPostprocesor(md), 'to_xml', 0)

# TODO: Create Leiden+ escape possibility that works with inline escapes `<= Some content =>`{.leiden+} or blocks:
# ```leiden+
# <=
# Some content
# =>
# ```



import re
from xml.etree import ElementTree as etree

from markdown.inlinepatterns import InlineProcessor


CA_DOT  = r'((?:ca\.)?)'
CA      = r'((?:ca)?)'
NUM     = r'(\d+(?:-\d+)?|\?)'
LINE    = r'((?:lin)?)'

RE_CHARACTER_GAP = fr'\[{CA}\.{NUM}\]'
RE_LINE_GAP      = fr'lost\.{CA_DOT}{NUM}lin'
RE_SPACE         = fr'vac\.{CA_DOT}{NUM}{LINE}'


def _handle_ca(ca, el):
    if ca:
        el.set('precision', 'low')

def _handle_num(num, el):
    if num == '?':
        el.set('extent', 'unknown')
    elif '-' in num:
        num_range = [int(i) for i in num.split('-')]
        el.set('atLeast', str(min(num_range)))
        el.set('atMost', str(max(num_range)))
    else:
        el.set('quantity', num)


class CharacterGapProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        ca, num = m.groups()
        el = etree.Element('gap')
        _handle_num(num, el)
        el.set('unit', 'character')
        _handle_ca(ca, el)
        return el, m.start(), m.end()


class LineGapProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        ca, num = m.groups()
        el = etree.Element('gap')
        _handle_num(num, el)
        el.set('unit', 'line')
        _handle_ca(ca, el)
        return el, m.start(), m.end()


class SpaceProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        ca, num, lin = m.groups()
        el = etree.Element('space')
        _handle_num(num, el)
        el.set('unit', 'line' if lin else 'character')
        _handle_ca(ca, el)
        return el, m.start(), m.end()



import re
from xml.etree import ElementTree as etree

from markdown.preprocessors import Preprocessor
from markdown.blockprocessors import BlockProcessor
from markdown.treeprocessors import Treeprocessor

from .exceptions import LeidenPlusSyntaxError

class DivisionsPreproc(Preprocessor):

    RE_OPEN = re.compile(r'(?:<D=\.(.+?)\.([a-z]+)|<D=.(r|v|le)|(<=))\s*')
    RE_CLOSE = re.compile(r'=(D?)>\s*')

    def _find_first_mark(self, line):
        open_match = self.RE_OPEN.search(line)
        close_match = self.RE_CLOSE.search(line)
        if (not open_match) and (not close_match):
            return None, ''
        elif (not open_match) and close_match:
            return close_match, 'c'
        elif open_match and (not close_match):
            return open_match, 'o'
        else: # open_match and close_match
            if open_match.start() < close_match.start():
                return open_match, 'o'
            else: # open_match.start() > close_match.start()
                return close_match, 'c'

    def run(self, lines):
        new_lines = []
        div_ids = []
        div_counter = 1

        in_ab = False
        for i, line in enumerate(lines):
            match, mark_type = self._find_first_mark(line)
            while match:
                new_lines.append(line[:match.start()])
                if mark_type == 'o':
                    if in_ab:
                        raise LeidenPlusSyntaxError('<= ... => cannot be nested')
                    n, div_type, rv, ab = match.groups()
                    if rv:
                        new_line = f'{{DIV-OPEN id={div_counter} n={rv}}}'
                    elif ab:
                        in_ab = True
                        new_line = f'{{AB-OPEN id={div_counter}}}'
                    else:
                        new_line = f'{{DIV-OPEN id={div_counter} subtype={div_type} n={n}}}'
                    new_lines += [
                        '', new_line, ''
                    ]
                    div_ids.append(div_counter)
                    div_counter += 1
                elif mark_type == 'c':
                    div = match.group(1)
                    if div and in_ab:
                        raise LeidenPlusSyntaxError('Incorrect division syntax: <= ... =D>')
                    elif (not div) and (not in_ab):
                        raise LeidenPlusSyntaxError('Incorrect division syntax: <D= ... =>')
                    try:
                        new_lines += ['', f'{{{"DIV" if div else "AB"}-CLOSE id={div_ids[-1]}}}', '']
                    except IndexError:
                        i_start = i - 3 if i >= 3 else 0
                        traceback = '\n'.join(lines[i_start:i+1])
                        raise LeidenPlusSyntaxError(
                            f"Text has a '={div}>' too many, traceback:\n\n{traceback}"
                        )
                    else:
                        del div_ids[-1]
                        if not div:
                            in_ab = False
                line = line[match.end():]
                match, mark_type = self._find_first_mark(line)
            new_lines.append(line)
        if div_ids != []:
            raise LeidenPlusSyntaxError(
                "Text ended while some Leiden+ division marks have not been closed"
            )
        return new_lines

class DivisionMarkProcessor(BlockProcessor):

    RE_OPEN = re.compile(r'^\{(DIV-OPEN|AB-OPEN) (.+?)\}')
    RE_CLOSE = re.compile(r'^\{(DIV-CLOSE|AB-CLOSE) (.+?)\}')

    def test(self, parent, block):
        return bool(self.RE_OPEN.search(block)) or bool(self.RE_CLOSE.search(block))

    def run(self, parent, blocks):
        lines = blocks.pop(0).split('\n')
        if len(lines) > 1 and not all(re.search(r'\s*', line) for line in lines[1:]):
            blocks.insert(0, '\n'.join(lines[1:]))

        match = self.RE_OPEN.search(lines[0])
        if not match:
            match = self.RE_CLOSE.search(lines[0])
        el = etree.SubElement(parent, match.group(1))
        attribs = [attr.split('=', maxsplit = 1) for attr in match.group(2).split(' ')]
        [el.set(*attrib) for attrib in attribs]


class DivisionMarkTreeproc(Treeprocessor):

    def run(self, root):
        for tag in ['AB', 'DIV']:
            el_close = root.find(f'./{tag}-CLOSE')
            while el_close != None:
                ID = el_close.attrib['id']
                el_open = root.find(f'./{tag}-OPEN[@id="{ID}"]')
                if el_open.tail:
                    el_open.text = el_open.tail
                    el_open.tail = None
                i_open = root.getchildren().index(el_open)
                i_close = root.getchildren().index(el_close)
                for _ in range(i_open + 1, i_close):
                    # We don't have to increase the index since it shifts
                    # automatically when elements are deleted
                    root[i_open].append(root[i_open + 1])
                    del root[i_open + 1]
                el_open.tag = tag.lower()
                el_open.attrib.pop('id')
                if tag == 'DIV':
                    el_open.attrib['type'] = 'textpart'
                del root[root.getchildren().index(el_close)]
                el_close = root.find(f'./{tag}-CLOSE')

class TrivialProcessor(BlockProcessor):
    """ Fallback alternative to the ParagraphProcessor. """

    def test(self, parent, block):
        return True

    def run(self, parent, blocks):
        block = blocks.pop(0)
        content = block
        children = parent.getchildren()
        if len(children) == 0:
            if parent.text:
                parent.text += content
            else:
                parent.text = content
        else:
            last_child = children[-1]
            if last_child.tail:
                last_child.tail += content
            else:
                last_child.tail = content



import re
from xml.etree import ElementTree as etree

from markdown.inlinepatterns import InlineProcessor


RE_MIL = fr'<=\.(ms|lb|cb|pb|gb) ([^\n]+?)\s*=>'

class MilestoneProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        mil_type, content = m.groups()
        cont_parts = re.split(r'\s+', content)
        el = etree.Element('milestone')
        el.set('unit', 'section')

        if mil_type == 'ms':
            if len(cont_parts) == 1:
                if content.startswith('par.'):
                    el.set('type', 'paragraph')
                    el.set('n', content[4:])
                else:
                    el.set('n', content)
            elif len(cont_parts) == 2:
                if re.search('^[0-9.]+$', cont_parts[0]):
                    el.set('ed', cont_parts[1])
                    el.set('n', cont_parts[0])
                else:
                    el.set('ed', cont_parts[0])
                    el.set('n', cont_parts[1])
            else:
                el.set('ed', ' '.join(cont_parts[:-1]))
                el.set('n', cont_parts[-1])

        elif mil_type == 'pb':
            el.set('type', 'page')
            if len(cont_parts) == 1 and re.search(r'^.+\.page\..+$', content):
                cont_parts = content.split('.')
                el.set('ed', cont_parts[0])
                el.set('n', cont_parts[2])
            elif len(cont_parts) == 2:
                if re.search('^[0-9.]+$', cont_parts[0]):
                    el.set('ed', cont_parts[1])
                    el.set('n', cont_parts[0])
                else:
                    el.set('ed', cont_parts[0])
                    el.set('n', cont_parts[1])
            else:
                el.set('ed', cont_parts[:-1])
                el.set('n', cont_parts[-1])

        elif mil_type in ['lb', 'cb', 'gb']:
            el.set('type', {'lb':'line', 'cb':'column', 'gb':'gathering'}[mil_type])
            if len(cont_parts) > 1:
                el.set('ed', cont_parts[:-1])
            el.set('n', cont_parts[-1])

        return el, m.start(), m.end()



from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

setup(
    name = 'leidenmark',
    version = '0.1.4',
    description = 'A markdown extension for converting Leiden+ epigraphic text to TEI Epidoc XML',
    packages = find_packages(),
    install_requires = ['markdown', 'lxml'],
    keywords = 'Leiden+ Markdown TEI Epidoc XML HTML',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = "https://github.com/andredelft/leidenmark",
    python_requires = '>=3.6',
    author = 'André van Delft',
    author_email = 'andrevandelft@outlook.com',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3'
    ]
)
